from pathlib import Path

from pydantic import BaseModel, Field


class AppConfig(BaseModel):
    config_env_var: str = Field(
        default="OPEN_TICKET_AI_CONFIG",
        description="Environment variable name for configuration file path",
    )
    config_yaml_root_key: str = Field(
        default="open_ticket_ai",
        description="Root key in YAML configuration file",
    )
    default_config_filename: str = Field(
        default="config.yml",
        description="Default configuration filename",
    )

    def get_default_config_path(self) -> Path:
        return Path.cwd() / self.default_config_filename

    def get_templates_dir(self) -> Path:
        package_root = Path(__file__).parent.parent.parent.parent.parent
        return package_root / "docs" / "raw_en_docs" / "config_examples"



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any

from pydantic import BaseModel, ConfigDict

from ..config.renderable import Renderable
from ..logging_iface import LoggerFactory
from .pipe_config import PipeConfig, PipeResult
from .pipe_context import PipeContext


class ParamsModel(BaseModel):
    model_config = ConfigDict(extra="allow")


class Pipe[ParamsT: ParamsModel](Renderable, ABC):
    params_class: type[ParamsT]

    def __init__(
        self, pipe_params: PipeConfig[ParamsT], logger_factory: LoggerFactory, *args: Any, **kwargs: Any
    ) -> None:
        self.pipe_config = pipe_params
        self._logger = logger_factory.get_logger(self.__class__.__name__)

        if isinstance(pipe_params.params, dict):
            self.params: ParamsT = self.params_class.model_validate(pipe_params.params)
        else:
            self.params: ParamsT = pipe_params.params

    def _save_pipe_result(self, context: PipeContext, pipe_result: PipeResult[Any]) -> PipeContext:
        if self.pipe_config.id is not None:
            context.pipes[self.pipe_config.id] = pipe_result
        return context

    def have_dependent_pipes_been_run(self, context: PipeContext) -> bool:
        return all(context.has_succeeded(dependency_id) for dependency_id in self.pipe_config.depends_on)

    async def process(self, context: PipeContext) -> PipeContext:
        self._logger.info(f"Processing pipe '{self.pipe_config.id}'")
        if self.pipe_config.should_run and self.have_dependent_pipes_been_run(context):
            self._logger.info(f"Pipe '{self.pipe_config.id}' is running.")
            return await self.__process_and_save(context)
        self._logger.info(f"Skipping pipe '{self.pipe_config.id}'.")
        return context

    async def __process_and_save(self, context: PipeContext) -> PipeContext:
        new_context = context.model_copy()
        try:
            pipe_result = await self._process()
        except Exception as e:
            self._logger.error(f"Error in pipe {self.pipe_config.id}: {str(e)}", exc_info=True)
            pipe_result = PipeResult[ParamsModel](success=False, failed=True, message=str(e), data=ParamsModel())
        updated_context = self._save_pipe_result(new_context, pipe_result)
        return updated_context

    @abstractmethod
    async def _process(self) -> PipeResult[Any]:
        pass



from open_ticket_ai.core.template_rendering.renderer_config import (
    JinjaRendererConfig,
    TemplateRendererConfig,
    TemplateRendererEnvConfig,
)
from open_ticket_ai.core.template_rendering.template_renderer import TemplateRenderer

__all__ = [
    "TemplateRenderer",
    "JinjaRendererConfig",
    "TemplateRendererConfig",
    "TemplateRendererEnvConfig",
]



from __future__ import annotations

from pydantic import BaseModel


class UnifiedNote(BaseModel):
    id: str | None = None
    subject: str = ""
    body: str = ""


class UnifiedEntity(BaseModel):
    id: str | None = None
    name: str | None = None


class UnifiedTicketBase(BaseModel):
    id: str | None = None
    subject: str | None = None
    queue: UnifiedEntity | None = None
    priority: UnifiedEntity | None = None
    notes: list[UnifiedNote] | None = None


class UnifiedTicket(UnifiedTicketBase):
    body: str | None = None


class TicketSearchCriteria(BaseModel):
    queue: UnifiedEntity | None = None
    limit: int | None = 10
    offset: int | None = 0



import asyncio
from typing import Any

from pydantic import BaseModel

from open_ticket_ai.core.orchestration.orchestrator_config import TriggerDefinition
from open_ticket_ai.core.orchestration.trigger import Trigger


class IntervalTriggerParams(BaseModel):
    milliseconds: int = 0
    seconds: int = 0
    minutes: int = 0
    hours: int = 0
    days: int = 0


class IntervalTrigger(Trigger[IntervalTriggerParams]):
    def __init__(self, config: TriggerDefinition[IntervalTriggerParams], *args: Any, **kwargs: Any) -> None:
        super().__init__(config, *args, **kwargs)
        self.seconds_interval: float = (
            self.params.days * 86400
            + self.params.hours * 3600
            + self.params.minutes * 60
            + self.params.seconds
            + self.params.milliseconds / 1000
        )
        self._task: asyncio.Task[None] | None = None

    def start(self) -> None:
        if not self._running:
            self._running = True
            self._task = asyncio.create_task(self._run())

    def stop(self) -> None:
        self._running = False
        if self._task:
            self._task.cancel()
            self._task = None

    async def _run(self) -> None:
        while self._running:
            await asyncio.sleep(self.seconds_interval)
            self.notify()



from __future__ import annotations

import logging
from logging.config import dictConfig
from typing import Any

from open_ticket_ai.core.config.config_models import LoggingDictConfig
from open_ticket_ai.core.logging_iface import AppLogger, LoggerFactory


class StdlibLogger(AppLogger):
    def __init__(self, logger: logging.Logger, context: dict[str, Any] | None = None):
        self._logger = logger
        self._context = context or {}

    def bind(self, **kwargs: Any) -> AppLogger:
        new_context = {**self._context, **kwargs}
        return StdlibLogger(self._logger, new_context)

    def _format_message(self, message: str, **kwargs: Any) -> str:
        all_context = {**self._context, **kwargs}
        if all_context:
            context_str = " ".join(f"{k}={v}" for k, v in all_context.items())
            return f"{message} [{context_str}]"
        return message

    def debug(self, message: str, *args: Any, **kwargs: Any) -> None:
        if args:
            self._logger.debug(message, *args)
        else:
            self._logger.debug(self._format_message(message, **kwargs))

    def info(self, message: str, *args: Any, **kwargs: Any) -> None:
        if args:
            self._logger.info(message, *args)
        else:
            self._logger.info(self._format_message(message, **kwargs))

    def warning(self, message: str, *args: Any, **kwargs: Any) -> None:
        if args:
            self._logger.warning(message, *args)
        else:
            self._logger.warning(self._format_message(message, **kwargs))

    def error(self, message: str, *args: Any, **kwargs: Any) -> None:
        if args:
            self._logger.error(message, *args)
        else:
            self._logger.error(self._format_message(message, **kwargs))

    def exception(self, message: str, *args: Any, **kwargs: Any) -> None:
        if args:
            self._logger.exception(message, *args)
        else:
            self._logger.exception(self._format_message(message, **kwargs))


class StdlibLoggerFactory(LoggerFactory):
    def get_logger(self, name: str, **context: Any) -> AppLogger:
        logger = logging.getLogger(name)
        return StdlibLogger(logger, context)


def create_logger_factory(logging_config: LoggingDictConfig) -> LoggerFactory:
    print(logging_config.model_dump_json(indent=2, by_alias=True, exclude_none=True))
    dictConfig(logging_config.model_dump(by_alias=True, exclude_none=True))
    return StdlibLoggerFactory()



from __future__ import annotations

import typing
from pydoc import locate
from typing import Any

from injector import inject, singleton

from open_ticket_ai.core import AppConfig
from open_ticket_ai.core.config.renderable import Renderable, RenderableConfig
from open_ticket_ai.core.logging_iface import LoggerFactory
from open_ticket_ai.core.pipeline.pipe import ParamsModel, Pipe
from open_ticket_ai.core.pipeline.pipe_config import PipeConfig
from open_ticket_ai.core.pipeline.pipe_context import PipeContext
from open_ticket_ai.core.template_rendering.template_renderer import TemplateRenderer


def _locate(use: str) -> type:
    if ":" in use:
        m, c = use.split(":", 1)
        use = f"{m}.{c}"
    use_class = locate(use)
    if use_class is None:
        raise ValueError(f"Cannot locate class '{use}'")
    return typing.cast(type, locate(use))


def render_params(params: ParamsModel, scope: PipeContext, renderer: TemplateRenderer) -> dict[str, Any]:
    if isinstance(params, dict):
        return params

    try:
        params_dict = params.model_dump()
    except Exception:
        params_dict = {}
        if hasattr(params, "__dict__"):
            for key, value in params.__dict__.items():
                if not key.startswith("_"):
                    params_dict[key] = value

    return renderer.render_recursive(params_dict, scope)


@singleton
class RenderableFactory:
    @inject
    def __init__(
        self,
        template_renderer: TemplateRenderer,
        app_config: AppConfig,
        registerable_configs: list[RenderableConfig[ParamsModel]],
        logger_factory: LoggerFactory,
    ):
        self._logger = logger_factory.get_logger(self.__class__.__name__)
        self._template_renderer = template_renderer
        self._registerable_configs = registerable_configs
        self._app_config = app_config
        self._logger_factory = logger_factory

    def create_pipe(self, pipe_config_raw: PipeConfig[ParamsModel], scope: PipeContext) -> Pipe[Any]:
        self._logger.debug(f"Creating pipe with config id: {pipe_config_raw.id}")
        self._logger.info(f"Creating pipe '{pipe_config_raw.id}'")
        rendered_params = render_params(pipe_config_raw.params, scope, self._template_renderer)
        pipe_config_raw.params = rendered_params
        registerable = self.__create_renderable_instance(pipe_config_raw, scope)
        if not isinstance(registerable, Pipe):
            raise TypeError(f"Registerable with id '{pipe_config_raw.id}' is not a Pipe")
        return registerable

    def create_trigger(self, trigger_config_raw: RenderableConfig[ParamsModel], scope: PipeContext) -> Renderable:
        self._logger.debug(f"Creating trigger with config id: {trigger_config_raw.id}")
        self._logger.info(f"Creating trigger '{trigger_config_raw.id}'")
        rendered_params = render_params(trigger_config_raw.params, scope, self._template_renderer)
        trigger_config_raw.params = rendered_params
        return self.__create_renderable_instance(trigger_config_raw, scope)

    def __create_service_instance(
        self, registerable_config_raw: RenderableConfig[ParamsModel], scope: PipeContext
    ) -> Renderable:
        rendered_params = render_params(registerable_config_raw.params, scope, self._template_renderer)
        registerable_config_raw.params = rendered_params
        return self.__create_renderable_instance(registerable_config_raw, scope)

    def __create_renderable_instance(
        self, registerable_config: RenderableConfig[ParamsModel], scope: PipeContext
    ) -> Renderable:
        cls: type = _locate(registerable_config.use)
        if not issubclass(cls, Renderable):
            raise TypeError(f"Class '{registerable_config.use}' is not a {Renderable.__class__.__name__}")

        kwargs: dict[str, Any] = {}
        kwargs |= self.__resolve_injects(registerable_config.injects, scope)
        kwargs["factory"] = self
        kwargs["app_config"] = self._app_config
        kwargs["logger_factory"] = self._logger_factory
        kwargs["config"] = registerable_config
        kwargs["pipe_config"] = registerable_config
        return cls(**kwargs)

    def __resolve_injects(self, injects: dict[str, str], scope: PipeContext) -> dict[str, Renderable]:
        out: dict[str, Any] = {}
        for param, ref in injects.items():
            out[param] = self.__resolve_by_id(ref, scope)
        return out

    def __resolve_by_id(self, service_id: str, scope: PipeContext) -> Any:
        for definition_config_raw in self._registerable_configs:
            definition_config: RenderableConfig[Any] = RenderableConfig.model_validate(definition_config_raw)
            if definition_config.id != service_id:
                continue
            return self.__create_service_instance(definition_config_raw, scope)
        raise KeyError(service_id)



from __future__ import annotations

import asyncio
from typing import Any

from injector import inject, singleton

from open_ticket_ai.core.config.renderable_factory import RenderableFactory
from open_ticket_ai.core.logging_iface import LoggerFactory
from open_ticket_ai.core.orchestration.orchestrator_config import (
    OrchestratorConfig,
    TriggerDefinition,
)
from open_ticket_ai.core.orchestration.scheduled_runner import PipeRunner
from open_ticket_ai.core.orchestration.trigger import Trigger
from open_ticket_ai.core.pipeline.pipe_context import PipeContext


@singleton
class Orchestrator:
    """Manages pipeline execution using Observer Pattern for triggers."""

    @inject
    def __init__(
        self, pipe_factory: RenderableFactory, orchestrator_config: OrchestratorConfig, logger_factory: LoggerFactory
    ) -> None:
        self._pipe_factory = pipe_factory
        self._config = orchestrator_config
        self._logger = logger_factory.get_logger(self.__class__.__name__)
        self._logger_factory = logger_factory
        self._trigger_registry: dict[str, Trigger[Any]] = {}
        self._runners: dict[str, PipeRunner] = {}

    def _instantiate_trigger(self, trigger_def: TriggerDefinition[Any]) -> Trigger[Any]:
        """Instantiate trigger using RenderableFactory for consistency with pipe instantiation."""
        scope = PipeContext()
        trigger: Trigger[Any] = self._pipe_factory.create_trigger(trigger_def, scope)  # type: ignore[assignment]
        return trigger

    def start(self) -> None:
        """Start the orchestrator and all runners."""
        self._logger.info(f"Starting orchestrator with {len(self._config.runners)} runner(s)")

        for index, definition in enumerate(self._config.runners):
            runner = PipeRunner(definition, self._pipe_factory, self._logger_factory)
            job_id = f"{definition.pipe_id}_{index}"
            self._runners[job_id] = runner

            for _trigger_index, trigger_def in enumerate(definition.on):
                # Ensure trigger has a valid ID (never None)
                trigger_id = trigger_def.id
                if trigger_id is None:
                    raise ValueError(
                        f"Trigger in runner '{definition.pipe_id}' has no ID. "
                        f"All triggers must have a unique 'id' field."
                    )

                if trigger_id in self._trigger_registry:
                    trigger = self._trigger_registry[trigger_id]
                else:
                    trigger = self._instantiate_trigger(trigger_def)
                    self._trigger_registry[trigger_id] = trigger

                trigger.attach(runner)
                self._logger.info(
                    f"Attached pipe '{definition.pipe_id}' to trigger '{trigger_def.id}' ({trigger_def.use})"
                )

        # Start all triggers
        for trigger in self._trigger_registry.values():
            trigger.start()

        self._logger.info("Orchestrator started successfully")

    def stop(self) -> None:
        """Stop the orchestrator and all triggers."""
        self._logger.info("Stopping orchestrator")
        for trigger in self._trigger_registry.values():
            trigger.stop()
        self._trigger_registry.clear()
        self._runners.clear()
        self._logger.info("Orchestrator stopped successfully")

    async def run(self) -> None:
        """Start the orchestrator and keep it running. Blocks until shutdown."""
        self.start()

        try:
            await asyncio.Future()  # Run forever
        except (KeyboardInterrupt, SystemExit) as e:
            self._logger.info(f"{e.__class__.__name__} received; shutting down orchestrator")
            self.stop()



from typing import Any

from pydantic import BaseModel

from open_ticket_ai.core.logging_iface import LoggerFactory
from open_ticket_ai.core.pipeline.pipe import ParamsModel, Pipe
from open_ticket_ai.core.pipeline.pipe_config import PipeConfig, PipeResult
from open_ticket_ai.core.ticket_system_integration.ticket_system_service import TicketSystemService
from open_ticket_ai.core.ticket_system_integration.unified_models import UnifiedNote


class AddNoteParams(ParamsModel):
    ticket_id: str | int
    note: UnifiedNote


class AddNotePipeResultData(BaseModel):
    note_added: bool


class AddNotePipeConfig(PipeConfig[AddNoteParams]):
    pass


class AddNotePipe(Pipe[AddNoteParams]):
    params_class = AddNoteParams

    def __init__(
        self,
        ticket_system: TicketSystemService,
        pipe_config: AddNotePipeConfig,
        logger_factory: LoggerFactory,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        super().__init__(pipe_config, logger_factory=logger_factory)
        self.pipe_config = AddNotePipeConfig.model_validate(pipe_config.model_dump())
        self.ticket_system = ticket_system

    async def _process(self) -> PipeResult[AddNotePipeResultData]:
        try:
            ticket_id_str = str(self.params.ticket_id)
            success = await self.ticket_system.add_note(ticket_id_str, self.params.note)
            if not success:
                return PipeResult[AddNotePipeResultData](
                    success=False,
                    failed=True,
                    message="Failed to add note to ticket",
                    data=AddNotePipeResultData(note_added=False),
                )
            return PipeResult[AddNotePipeResultData](
                success=True, failed=False, data=AddNotePipeResultData(note_added=True)
            )
        except Exception as e:
            return PipeResult[AddNotePipeResultData](
                success=False, failed=True, message=str(e), data=AddNotePipeResultData(note_added=False)
            )



from open_ticket_ai.core.config.app_config import AppConfig
from open_ticket_ai.core.config.config_loader import ConfigLoader
from open_ticket_ai.core.config.config_models import RawOpenTicketAIConfig
from open_ticket_ai.core.dependency_injection.container import AppModule

__all__ = [
    "AppConfig",
    "AppModule",
    "ConfigLoader",
    "RawOpenTicketAIConfig",
]



from typing import Annotated, Literal

from pydantic import BaseModel, ConfigDict, Field


class TemplateRendererEnvConfig(BaseModel):
    prefix: str | None = Field(default="OTAI_", description="Primary environment variable prefix")
    allowlist: set[str] | None = Field(default=None, description="Allowed environment variable names")
    denylist: set[str] | None = Field(default=None, description="Denied environment variable names")


class TemplateRendererConfig(BaseModel):
    model_config = ConfigDict(extra="allow")
    type: str = Field(..., description="Type of template renderer")
    env_config: TemplateRendererEnvConfig = Field(
        default_factory=TemplateRendererEnvConfig, description="Environment variable configuration"
    )


class JinjaRendererConfig(TemplateRendererConfig):
    type: Literal["jinja"] = "jinja"
    autoescape: bool = Field(default=False, description="Enable autoescaping in Jinja2")
    trim_blocks: bool = Field(default=True, description="Trim blocks in Jinja2")
    lstrip_blocks: bool = Field(default=True, description="Left-strip blocks in Jinja2")


class MustacheRendererConfig(TemplateRendererConfig):
    type: Literal["mustache"] = "mustache"


SpecificTemplateRendererConfig = Annotated[JinjaRendererConfig | MustacheRendererConfig, Field(discriminator="type")]



import os
from logging.config import dictConfig

from injector import Binder, Module, multiprovider, provider, singleton

from open_ticket_ai.base.loggers.stdlib_logging_adapter import create_logger_factory
from open_ticket_ai.core import AppConfig
from open_ticket_ai.core.config.config_loader import ConfigLoader
from open_ticket_ai.core.config.config_models import (
    RawOpenTicketAIConfig,
)
from open_ticket_ai.core.config.renderable import RenderableConfig
from open_ticket_ai.core.config.renderable_factory import RenderableFactory, _locate
from open_ticket_ai.core.logging_iface import LoggerFactory
from open_ticket_ai.core.orchestration.orchestrator_config import OrchestratorConfig
from open_ticket_ai.core.pipeline.pipe import ParamsModel
from open_ticket_ai.core.template_rendering import JinjaRendererConfig
from open_ticket_ai.core.template_rendering.template_renderer import TemplateRenderer


class AppModule(Module):
    def __init__(self, config_path: str | os.PathLike[str] | None = None, app_config: AppConfig | None = None) -> None:
        """Initialize AppModule with optional config path.

        Args:
            config_path: Path to config.yml. If None, uses OPEN_TICKET_AI_CONFIG
                        environment variable or falls back to default location.
        """
        self.config_path = config_path
        self.app_config = app_config or AppConfig()

    def configure(self, binder: Binder) -> None:
        binder.bind(AppConfig, to=self.app_config, scope=singleton)
        # Create a temporary logger factory for config loading
        from open_ticket_ai.core.config.config_models import LoggingDictConfig

        temp_logger_factory = create_logger_factory(LoggingDictConfig())
        config_loader = ConfigLoader(self.app_config, temp_logger_factory)
        config = config_loader.load_config(self.config_path)
        print(config.infrastructure.logging.model_dump_json(indent=4, by_alias=True, exclude_none=True))
        dictConfig(config.infrastructure.logging.model_dump(by_alias=True, exclude_none=True))
        binder.bind(RawOpenTicketAIConfig, to=config, scope=singleton)
        binder.bind(RenderableFactory, scope=singleton)

    @provider
    def _create_renderer_from_service(
        self, config: RawOpenTicketAIConfig, logger_factory: LoggerFactory
    ) -> TemplateRenderer:
        service_id = config.infrastructure.default_template_renderer
        service_config = next((s for s in config.services if s.id == service_id), None)
        if not service_config:
            raise ValueError(f"Template renderer service with id '{service_id}' not found")

        cls = _locate(service_config.use)
        params = service_config.params.model_dump()
        config_obj = JinjaRendererConfig.model_validate(params)
        return cls(config_obj, logger_factory=logger_factory)  # type: ignore[abstract]

    @provider
    @singleton
    def provide_logger_factory(self, config: RawOpenTicketAIConfig) -> LoggerFactory:
        return create_logger_factory(config.infrastructure.logging)

    @provider
    def provide_orchestrator_config(self, config: RawOpenTicketAIConfig) -> OrchestratorConfig:
        return config.orchestrator

    @multiprovider
    def provide_registerable_configs(self, config: RawOpenTicketAIConfig) -> list[RenderableConfig[ParamsModel]]:
        return config.services



from __future__ import annotations

from collections.abc import Iterable
from functools import reduce
from typing import Any, Self

from pydantic import BaseModel, ConfigDict, Field

from open_ticket_ai.core.config.renderable import RenderableConfig


class PipeConfig[ParamsT: BaseModel](RenderableConfig[ParamsT]):
    model_config = ConfigDict(populate_by_name=True)
    if_: str | bool = Field(default="True", alias="if")
    depends_on: str | list[str] = []
    steps: list[Any] | None = None

    @property
    def should_run(self) -> str | bool:
        return self.if_


class CompositePipeResultData(BaseModel):
    model_config = ConfigDict(extra="allow")


class PipeResult[DataT: BaseModel](BaseModel):
    model_config = ConfigDict(extra="forbid")
    success: bool
    failed: bool
    message: str = ""
    data: DataT

    def __and__(self, other: Self) -> PipeResult[CompositePipeResultData]:
        merged_data_dict: dict[str, Any] = {**self.data.model_dump(), **other.data.model_dump()}
        merged_data = CompositePipeResultData.model_validate(merged_data_dict)
        merged_msg = ";\n ".join([m for m in [self.message, other.message] if m])
        return PipeResult[CompositePipeResultData](
            success=self.success and other.success,
            failed=self.failed and other.failed,
            message=merged_msg,
            data=merged_data,
        )

    @classmethod
    def union(cls, results: Iterable[PipeResult[DataT]]) -> PipeResult[CompositePipeResultData]:
        if not results:
            return PipeResult[CompositePipeResultData](success=True, failed=False, data=CompositePipeResultData())
        return reduce(lambda a, b: a & b, results)  # type: ignore[arg-type]



class OpenTicketAIError(Exception):
    """Base exception for all custom exceptions in the Open Ticket AI application."""

    pass


class ConfigurationError(OpenTicketAIError):
    """Base exception for configuration-related errors."""

    pass


class MissingConfigurationError(ConfigurationError):
    """Raised when a required configuration is missing."""

    pass


class InvalidConfigurationError(ConfigurationError):
    """Raised when a configuration value is invalid."""

    pass


class EnvironmentVariableError(ConfigurationError):
    """Raised when there's an issue with environment variables."""

    pass


class PipelineError(OpenTicketAIError):
    """Base exception for pipeline-related errors."""

    pass


class PipelineExecutionError(PipelineError):
    """Raised when there's an error during pipeline execution."""

    pass


class PipelineValidationError(PipelineError):
    """Raised when pipeline validation fails."""

    pass


class PipelineTimeoutError(PipelineError):
    """Raised when a pipeline operation times out."""

    pass


class TemplateError(OpenTicketAIError):
    """Base exception for template-related errors."""

    pass


class TemplateSyntaxError(TemplateError):
    """Raised when there's a syntax error in a template."""

    pass


class TemplateRenderingError(TemplateError):
    """Raised when there's an error rendering a template."""

    pass


class UndefinedVariableError(TemplateError):
    """Raised when an undefined variable is accessed in a template."""

    pass


class DataProcessingError(OpenTicketAIError):
    """Base exception for data processing errors."""

    pass


class DataValidationError(DataProcessingError):
    """Raised when data validation fails."""

    pass


class DataTransformationError(DataProcessingError):
    """Raised when data transformation fails."""

    pass


class APIError(OpenTicketAIError):
    """Base exception for API-related errors."""

    pass


class APIConnectionError(APIError):
    """Raised when there's a connection error with an external API."""

    pass


class APIResponseError(APIError):
    """Raised when an API returns an error response."""

    pass


class APIRateLimitExceeded(APIError):
    """Raised when API rate limits are exceeded."""

    pass


class AuthenticationError(OpenTicketAIError):
    """Base exception for authentication errors."""

    pass


class UnauthorizedError(AuthenticationError):
    """Raised when a user is not authorized to perform an action."""

    pass


class InvalidCredentialsError(AuthenticationError):
    """Raised when invalid credentials are provided."""

    pass


class FileOperationError(OpenTicketAIError):
    """Base exception for file operation errors."""

    pass


class FileNotFoundError(FileOperationError):
    """Raised when a file is not found."""

    pass


class FilePermissionError(FileOperationError):
    """Raised when there's a permission error with file operations."""

    pass


class DatabaseError(OpenTicketAIError):
    """Base exception for database-related errors."""

    pass


class DatabaseConnectionError(DatabaseError):
    """Raised when there's an error connecting to the database."""

    pass


class DatabaseQueryError(DatabaseError):
    """Raised when there's an error executing a database query."""

    pass


class ValidationError(OpenTicketAIError):
    """Base exception for validation errors."""

    pass


class InvalidInputError(ValidationError):
    """Raised when input data is invalid."""

    pass


class MissingRequiredFieldError(ValidationError):
    """Raised when a required field is missing."""

    pass


class AIProcessingError(OpenTicketAIError):
    """Base exception for AI/ML processing errors."""

    pass


class ModelLoadingError(AIProcessingError):
    """Raised when there's an error loading an AI model."""

    pass


class PredictionError(AIProcessingError):
    """Raised when there's an error during model prediction."""

    pass


class NetworkError(OpenTicketAIError):
    """Base exception for network-related errors."""

    pass


class ConnectionTimeoutError(NetworkError):
    """Raised when a network connection times out."""

    pass


class ServiceUnavailableError(NetworkError):
    """Raised when a required service is unavailable."""

    pass



from __future__ import annotations

from open_ticket_ai.core.config.renderable_factory import RenderableFactory
from open_ticket_ai.core.logging_iface import LoggerFactory
from open_ticket_ai.core.orchestration.orchestrator_config import RunnerDefinition
from open_ticket_ai.core.pipeline.pipe import Pipe
from open_ticket_ai.core.pipeline.pipe_context import PipeContext


class PipeRunner:
    def __init__(
        self, definition: RunnerDefinition, pipe_factory: RenderableFactory, logger_factory: LoggerFactory
    ) -> None:
        self.definition = definition
        self.pipe_factory = pipe_factory
        self._logger = logger_factory.get_logger(f"{self.__class__.__name__}.{definition.pipe_id}")

    async def on_trigger_fired(self) -> None:
        await self.execute()

    async def execute(self) -> None:
        self._logger.info("Executing pipe '%s'", self.definition.pipe_id)
        try:
            pipe = self.pipe_factory.create_pipe(
                pipe_config_raw=self.definition.run,
                scope=PipeContext(params=self.definition.run.model_dump()),
            )
            if pipe is None:
                self._logger.error("Failed to create pipe '%s'", self.definition.pipe_id)
                return
            if not isinstance(pipe, Pipe):
                self._logger.error("Created object is not a Pipe instance: %s", type(pipe))
                return

            context_result = await pipe.process(PipeContext())
            pipe_result = context_result.pipes.get(self.definition.pipe_id)

            if pipe_result and pipe_result.success:
                self._logger.info("Pipe '%s' completed successfully", self.definition.pipe_id)
            else:
                failure_message = pipe_result.message if pipe_result else "No result available"
                self._logger.warning(
                    "Pipe '%s' completed with failure: %s",
                    self.definition.pipe_id,
                    failure_message,
                )
        except Exception:
            self._logger.exception(
                "Pipe '%s' execution failed with exception",
                self.definition.pipe_id,
            )



from typing import Any

from injector import inject
from jinja2.sandbox import SandboxedEnvironment

from open_ticket_ai.base.template_renderers.jinja_renderer_extras import (
    at_path,
    build_filtered_env,
    has_failed,
    pipe_result,
)
from open_ticket_ai.core.logging_iface import LoggerFactory
from open_ticket_ai.core.template_rendering.renderer_config import JinjaRendererConfig
from open_ticket_ai.core.template_rendering.template_renderer import TemplateRenderer


class JinjaRenderer(TemplateRenderer):
    @inject
    def __init__(self, config: JinjaRendererConfig, logger_factory: LoggerFactory):
        super().__init__(logger_factory)
        self.config = config
        self.jinja_env = SandboxedEnvironment(
            autoescape=config.autoescape, trim_blocks=config.trim_blocks, lstrip_blocks=config.lstrip_blocks
        )

    def render(self, template_str: str, scope: dict[str, Any], fail_silently: bool = False) -> Any:
        self.jinja_env.globals.update(scope)
        self.jinja_env.globals["at_path"] = at_path
        self.jinja_env.globals["env"] = build_filtered_env(self.config)
        self.jinja_env.globals["has_failed"] = has_failed
        self.jinja_env.globals["pipe_result"] = pipe_result
        try:
            template = self.jinja_env.from_string(template_str)
            rendered = template.render(self._to_dict(scope))
            return self._parse_rendered_value(rendered)
        except Exception as e:
            self._logger.warning("Failed to render template '%s'", template_str)
            self._logger.warning("context: %s", scope)
            self._logger.exception("Template rendering failed")
            if fail_silently:
                return template_str
            raise RuntimeError(e) from e



import asyncio
from abc import ABC, abstractmethod
from typing import Any, Protocol

from pydantic import BaseModel

from open_ticket_ai.core.config.renderable import Renderable
from open_ticket_ai.core.logging_iface import LoggerFactory
from open_ticket_ai.core.orchestration.orchestrator_config import TriggerDefinition


class PipeRunnerObserver(Protocol):
    async def on_trigger_fired(self) -> None: ...


class Trigger[ParamsT: BaseModel](Renderable, ABC):
    params_class: type[ParamsT]

    def __init__(
        self,
        config: TriggerDefinition[ParamsT],
        logger_factory: LoggerFactory,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        self.trigger_def = config
        self._observers: list[PipeRunnerObserver] = []
        self._running = False
        self._logger = logger_factory.get_logger(self.__class__.__name__)

        if isinstance(config.params, dict):
            self.params: ParamsT = self.params_class.model_validate(config.params)
        else:
            self.params: ParamsT = config.params

    def attach(self, observer: PipeRunnerObserver) -> None:
        if observer not in self._observers:
            self._observers.append(observer)

    def detach(self, observer: PipeRunnerObserver) -> None:
        if observer in self._observers:
            self._observers.remove(observer)

    def notify(self) -> None:
        for obs in self._observers:
            asyncio.create_task(obs.on_trigger_fired())

    @abstractmethod
    def start(self) -> None:
        pass

    @abstractmethod
    def stop(self) -> None:
        pass



from __future__ import annotations

from typing import Literal

from pydantic import BaseModel, ConfigDict, Field

from open_ticket_ai.core.config.renderable import EmptyParams, RenderableConfig
from open_ticket_ai.core.orchestration.orchestrator_config import OrchestratorConfig

LogLevel = Literal["NOTSET", "DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]


class FormatterConfig(BaseModel):
    model_config = ConfigDict(extra="allow", populate_by_name=True)
    class_: str | None = Field(default=None, alias="class")
    format: str | None = None
    datefmt: str | None = None
    style: Literal["%", "{", "$"] | None = None
    call: str | None = Field(default=None, alias="()")


class FilterConfig(BaseModel):
    model_config = ConfigDict(extra="allow", populate_by_name=True)
    class_: str | None = Field(default=None, alias="class")
    name: str | None = None
    call: str | None = Field(default=None, alias="()")


class HandlerConfig(BaseModel):
    model_config = ConfigDict(extra="allow", populate_by_name=True)
    class_: str = Field(alias="class")
    level: LogLevel | None = None
    formatter: str | None = None
    filters: list[str] | None = None
    call: str | None = Field(default=None, alias="()")


class LoggerConfig(BaseModel):
    model_config = ConfigDict(extra="allow")
    level: LogLevel | None = None
    handlers: list[str] | None = None
    propagate: bool | None = None
    filters: list[str] | None = None


class RootConfig(BaseModel):
    model_config = ConfigDict(extra="allow")
    level: LogLevel | None = None
    handlers: list[str] | None = None
    filters: list[str] | None = None


class LoggingDictConfig(BaseModel):
    model_config = ConfigDict(extra="allow")
    version: Literal[1] = 1
    disable_existing_loggers: bool | None = None
    incremental: bool | None = None
    root: RootConfig | None = None
    loggers: dict[str, LoggerConfig] = Field(default_factory=lambda: {})
    handlers: dict[str, HandlerConfig] = Field(default_factory=lambda: {})
    formatters: dict[str, FormatterConfig] = Field(default_factory=lambda: {})
    filters: dict[str, FilterConfig] = Field(default_factory=lambda: {})


class InfrastructureConfig(BaseModel):
    model_config = ConfigDict(extra="allow")
    logging: LoggingDictConfig = Field(default_factory=LoggingDictConfig)
    default_template_renderer: str


class RawOpenTicketAIConfig(BaseModel):
    model_config = ConfigDict(extra="allow")
    plugins: list[str] = Field(default_factory=lambda: [])
    infrastructure: InfrastructureConfig = Field(default_factory=InfrastructureConfig)
    services: list[RenderableConfig[EmptyParams]] = Field(default_factory=lambda: [])
    orchestrator: OrchestratorConfig = Field(default_factory=OrchestratorConfig)



from typing import Any

from pydantic import BaseModel

from open_ticket_ai.core.config.renderable_factory import RenderableFactory
from open_ticket_ai.core.logging_iface import LoggerFactory
from open_ticket_ai.core.pipeline.pipe import Pipe
from open_ticket_ai.core.pipeline.pipe_config import CompositePipeResultData, PipeConfig, PipeResult
from open_ticket_ai.core.pipeline.pipe_context import PipeContext


class CompositeParams(BaseModel):
    pass


class CompositePipeConfig(PipeConfig[CompositeParams]):
    steps: list[PipeConfig[Any]]


class CompositePipe(Pipe[CompositeParams]):
    """
    Composite pipe that runs multiple steps. Returns PipeResult from _process, Context from process.
    """

    params_class = CompositeParams

    def __init__(
        self,
        pipe_config: CompositePipeConfig,
        factory: RenderableFactory | None = None,
        logger_factory: LoggerFactory | None = None,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        super().__init__(pipe_config, logger_factory=logger_factory)
        self.pipe_config = CompositePipeConfig.model_validate(pipe_config.model_dump())
        self._factory = factory
        self._context: PipeContext | None = None

    def _build_pipe_from_step_config(self, step_config: PipeConfig[Any], context: PipeContext) -> Pipe[Any]:
        """
        Build a child pipe from step config.
        Returns Pipe.
        """
        if self._factory is None:
            raise ValueError("RenderableFactory is required but not provided to CompositePipe")
        return self._factory.create_pipe(step_config, context)

    async def _process_steps(self, context: PipeContext) -> list[PipeResult[Any]]:
        """
        Run all steps and collect their PipeResults.
        Returns list[PipeResult].
        """
        results: list[PipeResult[Any]] = []
        current_context = context
        for step_pipe_config_raw in self.pipe_config.steps or []:
            current_context.parent = context
            step_pipe = self._build_pipe_from_step_config(step_pipe_config_raw, current_context)
            current_context = await step_pipe.process(current_context)
            if step_pipe_config_raw.id in current_context.pipes:
                results.append(current_context.pipes[step_pipe_config_raw.id])
        self._context = current_context
        return results

    async def _process(self) -> PipeResult[CompositePipeResultData]:
        raise NotImplementedError("CompositePipe must override process() to access context")

    async def process(self, context: PipeContext) -> PipeContext:
        self._logger.info(f"Processing pipe '{self.pipe_config.id}'")
        if self.pipe_config.should_run and self.have_dependent_pipes_been_run(context):
            self._logger.info(f"Pipe '{self.pipe_config.id}' is running.")
            new_context = context.model_copy()
            try:
                steps_result: list[PipeResult[Any]] = await self._process_steps(new_context)
                composite_result = PipeResult.union(steps_result)
                if self._context:
                    new_context = self._context.model_copy()
            except Exception as e:
                self._logger.error(f"Error in pipe {self.pipe_config.id}: {str(e)}", exc_info=True)
                composite_result = PipeResult[CompositePipeResultData](
                    success=False, failed=True, message=str(e), data=CompositePipeResultData()
                )
            return self._save_pipe_result(new_context, composite_result)
        self._logger.info(f"Skipping pipe '{self.pipe_config.id}'.")
        return context



import uuid
from typing import Any, Self

from pydantic import BaseModel, ConfigDict, Field, model_validator


class Renderable:
    """Marker interface for renderable configurations."""

    pass


class EmptyParams(BaseModel):
    model_config = ConfigDict(extra="allow")
    pass


class RenderableConfig[ParamsT: BaseModel](BaseModel):
    model_config = ConfigDict(extra="allow", arbitrary_types_allowed=True)
    uid: str = Field(default_factory=lambda: uuid.uuid4().hex)
    id: str | None = None
    use: str = Field(default="open_ticket_ai.base.CompositePipe")
    injects: dict[str, str] = Field(default_factory=dict)
    params: ParamsT | dict[str, Any] = Field(default_factory=EmptyParams)  # type: ignore[assignment]

    @model_validator(mode="after")
    def set_id_from_uid(self) -> Self:
        if self.id is None:
            self.id = self.uid
        return self

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, RenderableConfig):
            return self.uid == other.uid
        return False

    def __hash__(self) -> int:
        return hash(self.uid)



from open_ticket_ai.base.pipes.composite_pipe import CompositePipe
from open_ticket_ai.base.pipes.expression_pipe import ExpressionPipe
from open_ticket_ai.base.pipes.ticket_system_pipes import AddNotePipe
from open_ticket_ai.base.pipes.ticket_system_pipes.fetch_tickets_pipe import FetchTicketsPipe
from open_ticket_ai.base.pipes.ticket_system_pipes.update_ticket_pipe import UpdateTicketPipe

__all__ = [
    "CompositePipe",
    "ExpressionPipe",
    "AddNotePipe",
    "FetchTicketsPipe",
    "UpdateTicketPipe",
]



import tomllib
from pathlib import Path

import pyfiglet
from injector import inject

from open_ticket_ai.core.config.config_models import RawOpenTicketAIConfig
from open_ticket_ai.core.logging_iface import LoggerFactory
from open_ticket_ai.core.orchestration.orchestrator import Orchestrator


def get_project_info() -> dict[str, str]:
    pyproject_path = Path("../../pyproject.toml")

    defaults = {
        "name": "Open Ticket AI",
        "version": "unknown",
        "license": "Not specified",
        "author_name": "Unknown",
        "author_email": "Unknown",
        "homepage": "Not specified",
    }

    if not pyproject_path.exists():
        return defaults

    with open(pyproject_path, "rb") as f:
        data = tomllib.load(f)

    project_data = data.get("project", {})

    authors = project_data.get("authors", [{}])
    author_name = authors[0].get("name", defaults["author_name"])
    author_email = authors[0].get("email", defaults["author_email"])

    license_info = project_data.get("license", {}).get("text", defaults["license"])

    homepage = project_data.get("urls", {}).get("Homepage", defaults["homepage"])

    return {
        "name": project_data.get("name", defaults["name"]),
        "version": project_data.get("version", defaults["version"]),
        "license": license_info,
        "author_name": author_name,
        "author_email": author_email,
        "homepage": homepage,
    }


class OpenTicketAIApp:
    @inject
    def __init__(self, config: RawOpenTicketAIConfig, orchestrator: Orchestrator, logger_factory: LoggerFactory):
        self.config = config
        self.orchestrator = orchestrator

        self._logger = logger_factory.get_logger(self.__class__.__name__)
        self.print_info()

    def print_info(self) -> None:
        project_info = get_project_info()
        banner = pyfiglet.figlet_format(project_info["name"])
        print(banner)
        self._logger.info(f" Version: {project_info['version']}")
        self._logger.info(f" License: {project_info['license']}")
        self._logger.info(f" Author:  {project_info['author_name']} <{project_info['author_email']}>")
        self._logger.info(f" Website: {project_info['homepage']}\n")

    async def run(self) -> None:
        self._logger.info("🚀 Starting Open Ticket AI orchestration...")
        self._logger.info(f"📦 Loaded {len(self.config.services)} services")
        self._logger.info(f"🔧 Orchestrator has {len(self.config.orchestrator.runners)} runners\n")

        try:
            await self.orchestrator.run()
        except KeyboardInterrupt:
            self._logger.info("\n⚠️  Shutdown requested...")
            self.orchestrator.stop()

        self._logger.info("✅ Orchestration complete")



import asyncio

from injector import Injector

from open_ticket_ai.app import OpenTicketAIApp
from open_ticket_ai.core.dependency_injection.container import AppModule


def get_container(config_path: str | None = None) -> Injector:
    return Injector([AppModule(config_path)])


async def run(config_path: str | None = None) -> None:
    container = get_container(config_path)
    app = container.get(OpenTicketAIApp)
    await app.run()


if __name__ == "__main__":
    asyncio.run(run())



import json
from pathlib import Path
from typing import Any

from pydantic import BaseModel

from open_ticket_ai.core.config.config_models import RawOpenTicketAIConfig


class RootConfig(BaseModel):
    open_ticket_ai: RawOpenTicketAIConfig


def get_type_description(type_info: dict[str, Any]) -> str:
    """Convert JSON Schema type info to human-readable string."""
    if "type" in type_info:
        type_val = type_info["type"]
        if isinstance(type_val, list):
            return " or ".join(str(t) for t in type_val)
        return str(type_val)

    if "anyOf" in type_info:
        types = [get_type_description(t) for t in type_info["anyOf"]]
        return " or ".join(types)

    if "allOf" in type_info:
        return "object"

    if "$ref" in type_info:
        ref_name = type_info["$ref"].split("/")[-1]
        return f"[{ref_name}](#{ref_name.lower()})"

    return "any"


def format_default(default: Any) -> str:
    """Format default value for display."""
    if isinstance(default, str):
        return f'`"{default}"`'
    if isinstance(default, (list, dict)):
        return f"`{json.dumps(default)}`"
    return f"`{default}`"


def generate_property_table(properties: dict[str, Any], required: list[str], defs: dict[str, Any]) -> str:
    """Generate a Markdown table for properties."""
    if not properties:
        return "_No properties defined._\n"

    lines = [
        "| Field | Type | Required | Default | Description |",
        "|-------|------|----------|---------|-------------|",
    ]

    for prop_name, prop_info in properties.items():
        is_required = "✓" if prop_name in required else ""

        # Get type
        type_desc = get_type_description(prop_info)

        # Get default
        default = prop_info.get("default", "")
        if default:
            default = format_default(default)

        # Get description
        description = prop_info.get("description", "").replace("\n", " ")

        lines.append(f"| `{prop_name}` | {type_desc} | {is_required} | {default} | {description} |")

    return "\n".join(lines) + "\n"


def generate_model_docs(name: str, schema: dict[str, Any], defs: dict[str, Any], level: int = 2) -> str:
    """Generate documentation for a single model."""
    heading = "#" * level
    lines = [f"{heading} {name}\n"]

    if "description" in schema:
        lines.append(f"{schema['description']}\n")

    properties = schema.get("properties", {})
    required = schema.get("required", [])

    lines.append(generate_property_table(properties, required, defs))

    return "\n".join(lines) + "\n"


def generate_markdown_docs(schema: dict[str, Any]) -> str:
    """Generate complete Markdown documentation from JSON Schema."""
    lines = [
        "# Configuration Schema Reference\n",
        "_Auto-generated from Pydantic models_\n",
        "---\n",
    ]

    defs = schema.get("$defs", {})

    # Generate root config
    lines.append(generate_model_docs("Root Configuration", schema, defs, level=2))

    # Generate definitions
    if defs:
        lines.append("## Type Definitions\n")
        for def_name, def_schema in sorted(defs.items()):
            lines.append(generate_model_docs(def_name, def_schema, defs, level=3))

    return "\n".join(lines)


if __name__ == "__main__":
    """Generates JSON schema for RootConfig and writes it to config.schema.json."""
    schema: dict[str, Any] = RootConfig.model_json_schema()
    generate_markdown_docs(schema)
    with open(Path.cwd() / "CONFIG_SCHEMA.md", "w", encoding="utf-8") as f:
        f.write(generate_markdown_docs(schema))
    with open(Path.cwd() / "config.schema.json", "w", encoding="utf-8") as f:
        json.dump(schema, f, indent=2)



from __future__ import annotations

from typing import Any

from pydantic import BaseModel, ConfigDict, Field

from open_ticket_ai.core.config.renderable import RenderableConfig
from open_ticket_ai.core.pipeline.pipe_config import PipeConfig


class TriggerDefinition[TriggerDefinitionParamsT: BaseModel](RenderableConfig[TriggerDefinitionParamsT]):
    pass


class ConcurrencySettings(BaseModel):
    max_workers: int = Field(default=1, ge=1)
    when_exhausted: str = Field(default="wait")

    model_config = ConfigDict(populate_by_name=True)


class RetrySettings(BaseModel):
    attempts: int = Field(default=3, ge=1)
    delay: str = Field(default="5s")
    backoff_factor: float = Field(default=2.0, ge=1.0)
    max_delay: str = Field(default="30s")
    jitter: bool = Field(default=True)

    model_config = ConfigDict(populate_by_name=True)


class RunnerParams(BaseModel):
    concurrency: ConcurrencySettings | None = None
    retry: RetrySettings | None = None
    timeout: str | None = None
    retry_scope: str = Field(default="pipeline")
    priority: int = Field(default=10)

    model_config = ConfigDict(populate_by_name=True)


class RunnerDefinition(BaseModel):
    id: str | None = None
    on: list[TriggerDefinition[BaseModel]]
    run: PipeConfig[Any]
    params: RunnerParams = Field(default_factory=RunnerParams)

    model_config = ConfigDict(populate_by_name=True)

    @property
    def pipe_id(self) -> str:
        if self.id is not None:
            return self.id
        if self.run.id is not None:
            return self.run.id
        return ""


class OrchestratorConfig(BaseModel):
    model_config = ConfigDict(extra="allow")
    defaults: dict[str, Any] | None = None
    runners: list[RunnerDefinition] = Field(default_factory=list)



from __future__ import annotations

from typing import Any

from pydantic import BaseModel, ConfigDict, Field

from open_ticket_ai.core.pipeline.pipe_config import PipeResult


class PipeContext(BaseModel):
    model_config = ConfigDict(ser_json_inf_nan="constants")

    pipes: dict[str, PipeResult[Any]] = Field(default_factory=dict)
    params: dict[str, Any] = Field(default_factory=dict)
    parent: PipeContext | None = Field(default=None, exclude=True)

    def has_succeeded(self, pipe_id: str) -> bool:
        pipe_result = self.pipes.get(pipe_id)
        if pipe_result is None:
            return False
        return pipe_result.success and not pipe_result.failed



from typing import Any

from pydantic import BaseModel

from open_ticket_ai.core.logging_iface import LoggerFactory
from open_ticket_ai.core.pipeline.pipe import Pipe
from open_ticket_ai.core.pipeline.pipe_config import PipeConfig, PipeResult
from open_ticket_ai.core.ticket_system_integration.ticket_system_service import TicketSystemService
from open_ticket_ai.core.ticket_system_integration.unified_models import TicketSearchCriteria


class FetchTicketsParams(BaseModel):
    ticket_search_criteria: TicketSearchCriteria | None = None


class FetchTicketsPipeResultData(BaseModel):
    fetched_tickets: list[dict[str, Any]]


class FetchTicketsPipeConfig(PipeConfig[FetchTicketsParams]):
    pass


class FetchTicketsPipe(Pipe[FetchTicketsParams]):
    params_class = FetchTicketsParams

    def __init__(
        self,
        ticket_system: TicketSystemService,
        pipe_config: FetchTicketsPipeConfig,
        logger_factory: LoggerFactory | None = None,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        super().__init__(pipe_config, logger_factory=logger_factory)
        self.pipe_config = FetchTicketsPipeConfig.model_validate(pipe_config.model_dump())
        self.ticket_system = ticket_system

    async def _process(self) -> PipeResult[FetchTicketsPipeResultData]:
        try:
            search_criteria = self.params.ticket_search_criteria
            if search_criteria is None:
                search_criteria = TicketSearchCriteria()
            tickets = await self.ticket_system.find_tickets(search_criteria) or []
            return PipeResult[FetchTicketsPipeResultData](
                success=True,
                failed=False,
                data=FetchTicketsPipeResultData(fetched_tickets=[ticket for ticket in tickets]),
            )
        except Exception as e:
            return PipeResult[FetchTicketsPipeResultData](
                success=False, failed=True, message=str(e), data=FetchTicketsPipeResultData(fetched_tickets=[])
            )



from typing import Any

from pydantic import BaseModel

from open_ticket_ai.core.logging_iface import LoggerFactory
from open_ticket_ai.core.pipeline.pipe import Pipe
from open_ticket_ai.core.pipeline.pipe_config import PipeConfig, PipeResult


class ExpressionParams(BaseModel):
    expression: str


class ExpressionPipeResultData(BaseModel):
    value: Any


class ExpressionPipeConfig(PipeConfig[ExpressionParams]):
    pass


class ExpressionPipe(Pipe[ExpressionParams]):
    params_class = ExpressionParams

    def __init__(
        self,
        pipe_config: ExpressionPipeConfig,
        logger_factory: LoggerFactory | None = None,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        super().__init__(pipe_config, logger_factory=logger_factory)
        self.pipe_config = ExpressionPipeConfig.model_validate(pipe_config.model_dump())
        self.expression = self.params.expression

    async def _process(self) -> PipeResult[ExpressionPipeResultData]:
        return PipeResult[ExpressionPipeResultData](
            success=True, failed=False, data=ExpressionPipeResultData(value=self.expression)
        )



from open_ticket_ai.base.pipes.ticket_system_pipes.add_note_pipe import (
    AddNoteParams,
    AddNotePipe,
    AddNotePipeConfig,
    AddNotePipeResultData,
)
from open_ticket_ai.base.pipes.ticket_system_pipes.fetch_tickets_pipe import (
    FetchTicketsParams,
    FetchTicketsPipe,
    FetchTicketsPipeConfig,
    FetchTicketsPipeResultData,
)
from open_ticket_ai.base.pipes.ticket_system_pipes.update_ticket_pipe import (
    UpdateTicketParams,
    UpdateTicketPipe,
    UpdateTicketPipeConfig,
    UpdateTicketPipeResultData,
)

FetchTicketsPipeConfig.model_rebuild()
AddNotePipeConfig.model_rebuild()
UpdateTicketPipeConfig.model_rebuild()

__all__ = [
    "AddNotePipe",
    "AddNotePipeConfig",
    "AddNoteParams",
    "AddNotePipeResultData",
    "FetchTicketsPipe",
    "FetchTicketsPipeConfig",
    "FetchTicketsParams",
    "FetchTicketsPipeResultData",
    "UpdateTicketPipe",
    "UpdateTicketPipeConfig",
    "UpdateTicketParams",
    "UpdateTicketPipeResultData",
]



from __future__ import annotations

import abc
from typing import Any


class AppLogger(abc.ABC):
    @abc.abstractmethod
    def bind(self, **kwargs: Any) -> AppLogger: ...

    @abc.abstractmethod
    def debug(self, message: str, **kwargs: Any) -> None: ...

    @abc.abstractmethod
    def info(self, message: str, **kwargs: Any) -> None: ...

    @abc.abstractmethod
    def warning(self, message: str, **kwargs: Any) -> None: ...

    @abc.abstractmethod
    def error(self, message: str, **kwargs: Any) -> None: ...

    @abc.abstractmethod
    def exception(self, message: str, **kwargs: Any) -> None: ...


class LoggerFactory(abc.ABC):
    @abc.abstractmethod
    def get_logger(self, name: str, **context: Any) -> AppLogger:
        pass



import os
from pathlib import Path

import injector
import yaml
from injector import singleton

from open_ticket_ai.core import AppConfig
from open_ticket_ai.core.config.config_models import RawOpenTicketAIConfig
from open_ticket_ai.core.logging_iface import LoggerFactory


@singleton
class ConfigLoader:
    @injector.inject
    def __init__(self, app_config: AppConfig, logger_factory: LoggerFactory):
        self.app_config = app_config
        self._logger = logger_factory.get_logger(self.__class__.__name__)

    def load_config(self, config_path: str | os.PathLike[str] | None = None) -> RawOpenTicketAIConfig:
        config_path_resolved: str | os.PathLike[str]
        if config_path is None and os.getenv(self.app_config.config_env_var) is not None:
            config_path_str = os.getenv(self.app_config.config_env_var)
            if config_path_str is None:
                raise ValueError("Config path from environment variable is None")
            config_path_resolved = config_path_str
        elif config_path is None and os.getenv(self.app_config.config_env_var) is None:
            config_path_resolved = self.app_config.get_default_config_path()
            if not Path(config_path_resolved).exists():
                raise FileNotFoundError(
                    f"Config file not found at {config_path_resolved}."
                    f"To fix this error:"
                    f"Create a confi file at {config_path_resolved}"
                    f"or provide a valid config path "
                    f"or set the {self.app_config.config_env_var} environment variable."
                )
        elif config_path is not None:
            config_path_resolved = config_path
        else:
            raise ValueError("Config path is None and no environment variable set")

        if not os.path.exists(config_path_resolved):
            raise FileNotFoundError(
                f"Config file not found at {config_path_resolved}"
                f"you need to create a config file at this path"
                f"or change the environment variable {self.app_config.config_env_var}"
            )

        with open(config_path_resolved) as file:
            yaml_content = yaml.safe_load(file)
            if yaml_content is None or self.app_config.config_yaml_root_key not in yaml_content:
                raise ValueError(f"Config file must contain '{self.app_config.config_yaml_root_key}' root key")
            config_dict = yaml_content[self.app_config.config_yaml_root_key]
            raw_otai_config = RawOpenTicketAIConfig.model_validate(config_dict)
        self._logger.info(f"Loaded config from {config_path}")
        return raw_otai_config


def load_config(
    config_path: str | os.PathLike[str] | None = None, app_config: AppConfig | None = None
) -> RawOpenTicketAIConfig:
    if app_config is None:
        app_config = AppConfig()

    loader = ConfigLoader(app_config)
    return loader.load_config(config_path)



import ast
import json
from abc import ABC, abstractmethod
from typing import Any

from pydantic import BaseModel

from open_ticket_ai.core.logging_iface import LoggerFactory


class TemplateRenderer(ABC):
    def __init__(self, logger_factory: LoggerFactory) -> None:
        self._logger = logger_factory.get_logger(self.__class__.__name__)

    def _to_dict(self, scope: BaseModel | dict[str, Any]) -> dict[str, Any]:
        if isinstance(scope, BaseModel):
            return scope.model_dump()
        return scope

    def _parse_rendered_value(self, s: str) -> Any:
        if not s.strip():
            return s

        stripped = s.strip()

        try:
            return json.loads(s)
        except Exception as e:
            self._logger.debug(f"Failed to parse JSON: {str(e)}")

            if stripped.startswith(("[", "{", "(", '"', "'")):
                try:
                    return ast.literal_eval(s)
                except Exception as e:
                    self._logger.debug(f"Failed to parse literal: {str(e)}")

            return s

    @abstractmethod
    def render(self, template_str: str, scope: dict[str, Any], fail_silently: bool = False) -> Any:
        pass

    def render_recursive(self, obj: Any, scope: BaseModel | dict[str, Any], fail_silently: bool = False) -> Any:
        self._logger.info(f"Rendering {obj}")
        self._logger.info(f"Scope: {scope}")
        scope_dict = self._to_dict(scope)

        if isinstance(obj, BaseModel):
            obj = self._to_dict(obj)
        if isinstance(obj, str):
            return self.render(obj, scope_dict, fail_silently)
        if isinstance(obj, list):
            return [self.render_recursive(item, scope_dict, fail_silently) for item in obj]
        if isinstance(obj, dict):
            return {k: self.render_recursive(v, scope_dict, fail_silently) for k, v in obj.items()}
        return obj



import ast
import json
import os
from typing import Any

from jinja2 import pass_context
from pydantic import BaseModel

from open_ticket_ai.core.pipeline.pipe_config import PipeResult
from open_ticket_ai.core.template_rendering.renderer_config import JinjaRendererConfig


def _coerce_path_to_list(path: str | list[str] | tuple[str, ...] | None) -> list[str]:
    if path is None:
        return []

    if isinstance(path, (list, tuple)):
        return [str(p) for p in path if str(p)]

    if not isinstance(path, str):
        return [str(path)]

    p = path.strip()
    if not p:
        return []

    if p.startswith(("[", "(")) and p.endswith(("]", ")")):
        seq = _try_parse_literal(p)
        if seq is not None:
            return [str(x) for x in seq]

    return [seg for seg in p.split(".") if seg]


def _try_parse_literal(s: str) -> list[Any] | tuple[Any, ...] | None:
    try:
        seq = ast.literal_eval(s)
        if isinstance(seq, (list, tuple)):
            return seq
    except (ValueError, SyntaxError):
        pass
    return None


def _nest_value_at_path(parts: list[str], value: Any) -> dict[str, Any] | Any:
    if not parts:
        return value

    result: dict[str, Any] | Any = value
    for key in reversed(parts):
        result = {key: result}
    return result


def _serialize_to_json(obj: Any) -> str:
    try:
        return json.dumps(obj)
    except (TypeError, ValueError):
        return str(obj)


def at_path(value: Any, path: str | list[str] | tuple[str, ...] | None) -> str:
    if isinstance(value, BaseModel):
        value = value.model_dump()

    parts = _coerce_path_to_list(path)
    nested = _nest_value_at_path(parts, value)
    return _serialize_to_json(nested)


@pass_context
def has_failed(ctx: Any, pipe_id: str) -> bool:
    pipes = ctx.get("pipes", {})
    pipe = pipes.get(pipe_id)
    if pipe is None:
        return False
    return bool(pipe.failed or pipe.get("failed"))


@pass_context
def pipe_result(ctx: Any, pipe_id: str, data_key: str = "value") -> Any:
    pipes = ctx.get("pipes", {})
    pipe = pipes.get(pipe_id)
    if pipe is None:
        return None
    pipe_data = pipe.data if isinstance(pipe, PipeResult) else pipe.get("data")
    if isinstance(pipe_data, BaseModel):
        pipe_data_dict = pipe_data.model_dump()
        return pipe_data_dict.get(data_key)
    return pipe_data.get(data_key)


def build_filtered_env(jinja_renderer_config: JinjaRendererConfig) -> dict[str, str]:
    env_config = jinja_renderer_config.env_config

    def _pref(k: str) -> bool:
        return True if not env_config.prefix else k.startswith(env_config.prefix)

    out = {k: v for k, v in dict(os.environ).items() if _pref(k)}
    if env_config.allowlist is not None:
        out = {k: v for k, v in out.items() if k in env_config.allowlist}
    if env_config.denylist is not None:
        out = {k: v for k, v in out.items() if k not in env_config.denylist}
    return out
